/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse;

import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

/**
 * {@code DatastoreTransaction} is a built-in entity represents a Datastore 
 * transaction.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@Entity
public class DatastoreTransaction {

	@Id private final String id;
	
	private List<Operation> operations = new ArrayList<Operation>();
	
	/**
	 * Constructs a new {@code DatastoreTransaction} instance with random UUID 
	 * based transaction id.
	 */
	public DatastoreTransaction() {
		this(UUID.randomUUID().toString());
	}
	
	/**
	 * Constructs a new {@code DatastoreTransaction} instance with the specified 
	 * transaction id.
	 * 
	 * @param id The transaction id.
	 */
	public DatastoreTransaction(String id) {
		this.id = id;
	}

	/**
	 * Returns the transaction id.
	 * 
	 * @return The transaction id.
	 */
	public String getId() {
		return id;
	}
	
	/**
	 * Returns the transactional operations at this datastore transaction.
	 * 
	 * @return The transactional operations at this datastore transaction.
	 */
	public List<Operation> getOperations() {
		return operations;
	}
	
	/**
	 * Adds a transactional {@code Operation}.
	 * 
	 * @param operation A transactional {@code Operation} to be added to this 
	 * datastore transaction.
	 */
	public void add(Operation operation) {
		operations.add(operation);
	}
	
	/**
	 * Returns the next operation sequence in this datastore transaction.
	 * 
	 * @return The nexe operation sequence in this datastore transaction.
	 */
	public long getSequence() {
		return operations.size() + 1;
	}
	
}
