/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse;

/**
 * {@code Delete} is a command interface to delete entities match to the 
 * specified {@code Filter}s in Acid House's transactional session. This 
 * command execution returns {@code Void} (nothing). You can build &amp; execute 
 * this command as the following code: 
 * <pre>
 * import org.eiichiro.acidhouse.DatastoreSession;
 * import org.eiichiro.acidhouse.appengine.AppEngineDatastoreSession;
 * ...
 * 
 * // Create 'DatastoreSession' instance.
 * DatastoreSession session = new AppEngineDatastoreSession();
 * // Begin transaction.
 * session.begin();
 * // Get metamodel instance of 'Entity3' class.
 * Entity3_ entity3_ = Metamodels.metamodel(Entity3.class);
 * 
 * try {
 * 	// Delete all of 'Entity3' entities which their 'entity1.i' property value is 
 * 	// less than 15.
 * 	session.delete(Entity3.class)
 * 			.filter(entity3_.entity1.i.lessThan(15))
 * 			.execute();
 * 	// Commit transaction.
 * 	session.commit();
 * } catch (Exception e) {
 * 	session.rollback();
 * 	throw e;
 * } finally {
 * 	session.close();
 * }
 * </pre>
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public interface Delete<E> extends Command<Void> {

	/**
	 * Qualifies entities to be deleted with the specified {@code Filter}s.
	 * Each of the specified filters is combined with "logical and".
	 * 
	 * @param filters {@code Filter}s to qualify entities to be deleted.
	 * @return The {@code Delete} to delete entities qualified with the 
	 * specified {@code Filter}s.
	 */
	public Delete<E> filter(Filter<?>... filters);
	
}
