/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse;

/**
 * {@code Operation} is a built-in entity and represents a basic CRUD operation 
 * in Datastore.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@Entity
public class Operation {

	/**
	 * {@code Type} is the type of datastore operation.
	 * 
	 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
	 */
	public static enum Type {
		
		/** Get (Read) operation. */
		GET, 
		
		/** Put (Create) operation. */
		PUT, 
		
		/** Update operation. */
		UPDATE, 
		
		/** Delete operation. */
		DELETE
		
	}
	
	@Id private final long sequence;
	
	private final Type type;
	
	private final Object entity;
	
	/**
	 * Constructs a new {@code Operation} instance with the specified operation 
	 * sequence in transaction, operation type and entity which this operation 
	 * is applied to.
	 * 
	 * @param sequence The operation sequence in transaction.
	 * @param type The operation type.
	 * @param entity The entity which this operation is applied to.
	 */
	public Operation(long sequence, Type type, Object entity) {
		this.sequence = sequence;
		this.type = type;
		this.entity = entity;
	}
	
	/**
	 * Returns the operation sequence in transaction.
	 * 
	 * @return The operation sequence in transaction.
	 */
	public long getSequence() {
		return sequence;
	}

	/**
	 * Returns the operation type.
	 * 
	 * @return The operation type.
	 */
	public Type getType() {
		return type;
	}

	/**
	 * Returns the entity which this operation is applied to.
	 * 
	 * @return The entity which this operation is applied to.
	 */
	public Object getEntity() {
		return entity;
	}
	
}
