/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.appengine;

import java.util.ArrayList;
import java.util.ConcurrentModificationException;
import java.util.Date;
import java.util.List;
import java.util.logging.Logger;

import org.eiichiro.acidhouse.DatastoreTransaction;
import org.eiichiro.acidhouse.Entities;
import org.eiichiro.acidhouse.EntityExistsException;
import org.eiichiro.acidhouse.IndoubtException;
import org.eiichiro.acidhouse.Lock;
import org.eiichiro.acidhouse.Operation;
import org.eiichiro.acidhouse.ResourceManager;

import com.google.appengine.api.datastore.Entity;
import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.KeyFactory;
import com.google.appengine.api.datastore.Query;
import com.google.appengine.api.datastore.Transaction;

/**
 * {@code AppEngineResourceManager} is a App Engine Low-level Datastore API 
 * based {@code ResourceManager} implementation.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class AppEngineResourceManager implements ResourceManager<Transaction> {

	private final Logger logger = Logger.getLogger(getClass().getName());
	
	private final AppEngineDatastoreService datastore;
	
	private final Transaction transaction;
	
	private final DatastoreTransaction datastoreTransaction;
	
	private Object entity;
	
	/**
	 * Constructs a new {@code AppEngineResourceManager} with the specified 
	 * {@code AppEngineDatastoreService}.
	 * 
	 * @param datastoreService {@code AppEngineDatastoreService}.
	 */
	public AppEngineResourceManager(AppEngineDatastoreService datastoreService) {
		this.datastore = datastoreService;
		transaction = null;
		datastoreTransaction = null;
	}
	
	/**
	 * Constructs a new {@code AppEngineResourceManager} with the specified 
	 * {@code AppEngineDatastoreService}, Google App Engine Datastore 
	 * {@code Transaction} and Acid House {@code DatastoreTransaction}.
	 * 
	 * @param datastore {@code AppEngineDatastoreService}.
	 * @param transaction Google App Engine Datastore {@code Transaction}.
	 * @param datastoreTransaction Acid House {@code DatastoreTransaction}.
	 */
	public AppEngineResourceManager(AppEngineDatastoreService datastore,
			Transaction transaction, DatastoreTransaction datastoreTransaction) {
		this.datastore = datastore;
		this.transaction = transaction;
		this.datastoreTransaction = datastoreTransaction;
	}
	
	/**
	 * Gets the entity instance of the specified {@code Class} corresponding to 
	 * the specified id.
	 * This method implements "Consistent read". This method gets the entity as 
	 * the following procedure.
	 * <ul>
	 * <li>
	 * Determines if the entity is locked. If the entity has been locked, goes 
	 * to the next step. If the entity has not been locked, jumps to the last 
	 * step.
	 * </li>
	 * <li>
	 * Attempts to get {@code DatastoreTransaction} from {@code Lock}. If the 
	 * {@code DatastoreTransaction} is found, goes to the next step. If the 
	 * {@code DatastoreTransaction} is not found and transaction timeout has 
	 * been elapsed, this method considers the transaction had been failed 
	 * (rolled back), then deletes {@code DatastoreTransaction} from Google App 
	 * Engine Datastore and goes to the last step. If the {@code DatastoreTransaction} 
	 * is not found and transaction timeout has not been elapsed, this method 
	 * assumes the entity is being modified in transaction and throws 
	 * {@code ConcurrentModificationException}.
	 * </li>
	 * <li>
	 * Gets the list of {@code Operation} from {@code DatastoreTransaction} and 
	 * applies every operation to entities, then unlock them.
	 * </li>
	 * <li>
	 * Gets the entity from Google App Engine Datastore and checks the entity 
	 * is locked again. if the entity has been locked, this method throws 
	 * {@code ConcurrentModificationException}. Otherwise, this method returns 
	 * the entity.
	 * </li>
	 * </ul>
	 * 
	 * @param <E> The type of entity.
	 * @param clazz The {@code Class} of entity.
	 * @param id The id corresponding to the entity you attempt to get.
	 * @return The entity instance of the specified {@code Class} corresponding 
	 * to the specified id.
	 * @throws ConcurrentModificationException If the entity has been modified 
	 * by other transaction.
	 * @throws IndoubtException If the data consistency broken is detected when 
	 * the entity is get.
	 */
	@Override
	public <E> E get(Class<E> clazz, Object id) throws ConcurrentModificationException, IndoubtException {
		Key key = Keys.create(Translation.toKind(clazz), id);
		unlock(key);
		List<Entity> entities = null;
		
		if ((transaction == null)) {
			entities = datastore.query(new Query(key));
		} else {
			entities = datastore.query(transaction, new Query(key));
		}
		
		for (Entity entity : entities) {
			if (entity.getKind().equals(Translation.LOCK_KIND)) {
				throw new ConcurrentModificationException(
						"Entity corresponding to [" + key + "] is in transaction");
			}
		}

		E object = Translation.toObject(clazz, entities, datastore);
		entity = object;
		type = Operation.Type.GET;
		return object;
	}
	
	/** Google App Engine Datastore transaction timeout. */
	public static final long TIMEOUT = 30000;
	
	private void unlock(Key key) {
		List<Entity> entities = datastore.query(new Query(Translation.LOCK_KIND, key));
		
		// If the entity corresponding to the specified key hasn't been locked, 
		// this method ignores it.
		if (entities.size() == 0) {
			return;
		}
		
		Entity lockEntity = entities.get(0);
		Lock lock = Translation.toLock(lockEntity);
		
		// If Transaction entities are not found, this method assumes that the 
		// entity corresponding to the specified key is in transaction when 
		// current time within a transaction timeout. 
		// When current time has exceeded a transaction timeout, this method 
		// assumes that the another transaction has already failed and just 
		// unlocks the entity.
		Key transactionKey = KeyFactory.stringToKey(lock.getTransaction());
		List<Entity> transactionEntities = new ArrayList<Entity>();
		
		if (datastore.get(transactionKey) == null) {
			if (System.currentTimeMillis() - lock.getTimestamp().getTime() <= TIMEOUT) {
				throw new ConcurrentModificationException(
						"Entity corresponding to [" + key + "] is in transaction");
			} else {
				Transaction beginTransaction = datastore.beginTransaction();
				
				if (datastore.get(beginTransaction, lockEntity.getKey()) == null) {
					logger.info("Entity corresponding to lock [" + lock
							+ "] has been rollbacked by another transaction");
					transaction.rollback();
					return;
				}
				
				datastore.delete(beginTransaction, lockEntity.getKey());
				beginTransaction.commit();
				return;
			}
			
		} else {
			transactionEntities.addAll(datastore.query(new Query(transactionKey)));
		}
		
		DatastoreTransaction datastoreTransaction 
				= Translation.toDatastoreTransaction(transactionEntities, datastore);
		Key ancestor = Keys.getAncestor(transactionKey);
		
		Operation owner = null;
		Key ownerLockKey = null;
		int committed = 0;
		
		for (Operation operation : datastoreTransaction.getOperations()) {
			Key entityKey = Keys.create(
					Translation.toKind(operation.getEntity().getClass()), 
					Entities.getId(operation.getEntity()));
			Key lockKey = Keys.create(entityKey, Translation.LOCK_KIND, lock.getId());
			
			// The owner entity of transaction entities must be deleted after 
			// every operation is applied. Because this method cannot apply any 
			// operation to entity without transaction entities.
			if (entityKey.equals(ancestor)) {
				owner = operation;
				ownerLockKey = lockKey;
				continue;
			}
			
			List<Key> deletes = new ArrayList<Key>();
			deletes.add(lockKey);
			
			// Applies operation to entity and unlock it.
			try {
				apply(operation.getType(), operation.getEntity(), lockKey, deletes);
				committed++;
			} catch (Exception e) {
				throwIndoubtException(e, datastoreTransaction, committed);
			}
		}
		
		// At last, this method applies operation to the owner entity, unlock it 
		// and deletes transaction entities.
		if (owner != null) {
			List<Key> deletes = new ArrayList<Key>();
			deletes.add(ownerLockKey);
			
			for (Entity delete : transactionEntities) {
				deletes.add(delete.getKey());
			}
			
			try {
				apply(owner.getType(), owner.getEntity(), ownerLockKey, deletes);
				committed++;
			} catch (Exception e) {
				throwIndoubtException(e, datastoreTransaction, committed);
			}
		}
		
		// Done!
	}
	
	private void apply(Operation.Type type, Object entity, Key lock, List<Key> deletes) {
		Transaction transaction = datastore.beginTransaction();
		
		if (datastore.get(transaction, lock) == null) {
			logger.info("Entity corresponding to lock [" + lock
					+ "] has been applied by another transaction");
			transaction.rollback();
			return;
		}
		
		if (type != Operation.Type.DELETE) {
			datastore.put(transaction, Translation.toEntities(entity));
		} else {
			for (Entity e : Translation.toEntities(entity)) {
				deletes.add(e.getKey());
			}
		}
		
		datastore.delete(transaction, deletes);
		transaction.commit();
	}
	
	private void throwIndoubtException(Exception exception, DatastoreTransaction transaction, int committed) {
		logger.severe("Transaction [" 
				+ transaction.getId() 
				+ "]: Failed to apply transaction log; " 
				+ "DATA CONSISTENCY BROKEN! PLEASE RECOVERY IMMEDIATELY " 
				+ "WITH THE FOLLOWING TRANSACTION DUMP :-<");
		throw new IndoubtException(exception, transaction, committed);
	}

	/**
	 * Returns the entity this {@code AppEngineResourceManager} manages.
	 * 
	 * @return The entity this {@code AppEngineResourceManager} manages.
	 */
	@Override
	public Object getEntity() {
		return entity;
	}

	/**
	 * Returns the {@code Transaction} this {@code AppEngineResourceManager} 
	 * manages.
	 * 
	 * @return The {@code Transaction} this {@code AppEngineResourceManager} 
	 * manages.
	 */
	@Override
	public Transaction getTransaction() {
		return transaction;
	}

	/** Allocates lock for the managed entity. */
	@Override
	public void prepare() {
		List<Operation> operations = datastoreTransaction.getOperations();
		Object parent = operations.get(operations.size() - 1).getEntity();
		Lock lock = new Lock(datastoreTransaction.getId(), KeyFactory.createKeyString(
				Keys.create(Translation.toKind(parent.getClass()), Entities.getId(parent)), 
				Translation.DATASTORE_TRANSACTION_KIND, 
				datastoreTransaction.getId()), new Date());
		datastore.put(transaction, Translation.toEntity(lock, Keys.create(
				Translation.toKind(entity.getClass()), Entities.getId(entity))));
		transaction.commit();
		this.lock = lock;
	}
	
	/**
	 * Applies the transactional operation (Redo log) to the managed entity and 
	 * unlock it. 
	 */
	@Override
	public void commit() {
		List<Key> deletes = new ArrayList<Key>();
		deletes.add(Translation.toEntity(lock, Keys.create(
				Translation.toKind(entity.getClass()), 
				Entities.getId(entity))).getKey());
		Key transactionKey = KeyFactory.stringToKey(lock.getTransaction());
		
		if (Keys.getAncestor(transactionKey).equals(
				Keys.create(Translation.toKind(entity.getClass()), Entities.getId(entity)))) {
			deletes.add(transactionKey);
			
			for (Entity entity : datastore.query(
					new Query(transactionKey).setKeysOnly())) {
				deletes.add(entity.getKey());
			}
		}
		
		Key parent = Keys.create(Translation.toKind(entity.getClass()), Entities.getId(entity));
		apply(type, entity, Translation.toEntity(lock, parent).getKey(), deletes);
	}
	
	private Operation.Type type;
	
	private Lock lock;
	
	/**
	 * Puts the specified entity instance into the Google App Engine Datastore 
	 * newly.
	 * 
	 * @param entity An entity instance.
	 */
	@Override
	public void put(Object entity) {
		if (entity == null) {
			throw new IllegalArgumentException("'entity' must not be [" + entity + "]");
		}
		
		if (type != null) {
			throw new IllegalStateException("Operation [" + type + "] -> ["
					+ Operation.Type.PUT
					+ "] is not allowed: This operation must be first");
		}
		
		Key id = Keys.create(Translation.toKind(entity.getClass()), Entities.getId(entity));
		Entity e = datastore.get(transaction, id);
		
		if (e != null) {
			throw new EntityExistsException(id);
		}
		
		this.entity = entity;
		type = Operation.Type.PUT;
	}

	/**
	 * Applies the specified entity's update to the Google App Engine Datastore.
	 *  
	 * @param entity An entity instance to be updated.
	 */
	@Override
	public void update(Object entity) {
		if (entity == null) {
			throw new IllegalArgumentException("'entity' must not be [" + entity + "]");
		}
		
		if (type != Operation.Type.GET) {
			throw new IllegalStateException("Operation [" + type + "] -> ["
					+ Operation.Type.UPDATE
					+ "] is not allowed: previous operation must be ["
					+ Operation.Type.GET + "]");
		}
		
		type = Operation.Type.UPDATE;
	}
	
	/**
	 * Deletes the specified entity from the Google App Engine Datastore.
	 * 
	 * @param entity An entity instance to be updated.
	 */
	@Override
	public void delete(Object entity) {
		if (entity == null) {
			throw new IllegalArgumentException("'entity' must not be [" + entity + "]");
		}
		
		if (type != Operation.Type.GET) {
			throw new IllegalStateException("Operation [" + type + "] -> ["
					+ Operation.Type.DELETE
					+ "] is not allowed: previous operation must be ["
					+ Operation.Type.GET + "]");
		}
		
		type = Operation.Type.DELETE;
	}
	
}
