/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.appengine;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import org.eiichiro.acidhouse.Filter;
import org.eiichiro.acidhouse.GetList;
import org.eiichiro.acidhouse.Update;
import org.eiichiro.acidhouse.metamodel.Property;

/**
 * {@code AppEngineUpdate} is a App Engine Low-level Datastore API based 
 * implentation of {@code Update}.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class AppEngineUpdate<E> implements Update<E> {

	private Logger logger = Logger.getLogger(getClass().getName());
	
	private final Class<E> kind;
	
	private List<Filter<?>> filters = new ArrayList<Filter<?>>(0);
	
	private Map<Property<?, ?>, Object> properties = new HashMap<Property<?, ?>, Object>(0);
	
	private final AppEngineDatastoreSession session;
	
	/**
	 * Constructs a new {@code AppEngineUpdate} with the specified kind of 
	 * entity and {@code AppEngineDatastoreSession}.
	 * 
	 * @param kind The kind of entity to delete with this command.
	 * @param session {@code AppEngineDatastoreSession} instance for this command.
	 */
	public AppEngineUpdate(Class<E> kind, AppEngineDatastoreSession session) {
		this.kind = kind;
		this.session = session;
	}
	
	/**
	 * Qualifies entities to be updated with the specified {@code Filter}s.
	 * Each of the specified filters is combined with "logical and".
	 * 
	 * @param filters {@code Filter}s to qualify entities to be updated.
	 * @return The {@code Update} to update entities qualified with the 
	 * specified {@code Filter}s.
	 */
	@Override
	public Update<E> filter(Filter<?>... filters) {
		if (filters == null) {
			throw new IllegalArgumentException("'filters' must not be [" + filters + "]");
		}
		
		this.filters = Arrays.asList(filters);
		return this;
	}

	/**
	 * Qualifies property and value to be updated by this {@code Update}.
	 * 
	 * @param property The property to be updated by this {@code Update}.
	 * @param value The value to be updated by this {@code Update}.
	 * @return The {@code Update} to update the specified entity 
	 * properties.
	 */
	@Override
	public <T> Update<E> property(Property<?, T> property, T value) {
		if (property == null) {
			throw new IllegalArgumentException("'property' must not be [" + property + "]");
		}
		
		properties.put(property, value); 
		return this;
	}
	
	/**
	 * Executes {@code Update} with {@code AppEngineDatastoreSession}.
	 */
	@SuppressWarnings("unchecked")
	@Override
	public Void execute() {
		logger.fine("Building AppEngineUpdate command");
		
		if (properties.size() == 0) {
			throw new IllegalStateException("'properties' size must be greater than 0");
		}
		
		GetList<E> command = session.get(kind);
		
		if (filters != null) {
			command.filter(filters.toArray(new Filter<?>[] {}));
		}
		
		for (E entity : command.execute()) {
			Iterator<Property<?, ?>> iterator = properties.keySet().iterator();
			
			while (iterator.hasNext()) {
				Property<E, ?> property = (Property<E, ?>) iterator.next();
				property.setValue(entity, properties.get(property));
			}
			
			session.update(entity);
		}
		
		return null;
	}
	
}
