/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.appengine;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import com.google.appengine.api.datastore.Entity;
import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.KeyFactory;

/**
 * {@code Keys} are utility methods for Google App Engine Datastore {@code Key}-
 * related operations.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public abstract class Keys {

	private Keys() {}
	
	/**
	 * Creates {@code Key} with the specified kind name and id.
	 * @param kind The kind name of entity.
	 * @param id The id of entity.
	 * @return {@code Key} created with the specified kind name and id.
	 */
	public static Key create(String kind, Object id) {
		return create(null, kind, id);
	}
	
	/**
	 * Creates {@code Key} with the specified parent, kind name and id.
	 * 
	 * @param parent The {@code Key} of parent.
	 * @param kind The kind name of entity.
	 * @param id The id of entity.
	 * @return {@code Key} created with the specified parent, kind name and id.
	 */
	public static Key create(Key parent, String kind, Object id) {
		if (id instanceof String) {
			return (parent == null) ? KeyFactory.createKey(kind, (String) id) : KeyFactory.createKey(parent, kind, (String) id);
		} else if (id instanceof Long) {
			return (parent == null) ? KeyFactory.createKey(kind, (Long) id) : KeyFactory.createKey(parent, kind, (Long) id);
		} else if (id instanceof Key) {
			return (Key) id;
		} else {
			throw new IllegalArgumentException("'id' must be an instance of ["
					+ String.class.getName() + "] or [" + Long.class.getName()
					+ "] or [" + Key.class.getName() + "]");
		}
	}
	
	/**
	 * Gets ancestor {@code Key} from the specified {@code Key}.
	 * 
	 * @param key The {@code Key} of entity.
	 * @return Ancestor {@code Key} of the specified {@code Key}.
	 */
	public static Key getAncestor(Key key) {
		Key ancestor = key;
		
		while (ancestor.getParent() != null) {
			ancestor = ancestor.getParent();
		}
		
		return ancestor;
	}
	
	/**
	 * Gets the {@code Key} pedigree of the specified entity.
	 * The ancestor is first entry.
	 * 
	 * @param entity Google App Engine Datastore 
	 * @return The {@code Key} pedigree of the specified entity.
	 */
	public static List<Key> getPedigree(Entity entity) {
		List<Key> pedigree = new ArrayList<Key>();
		Key key = entity.getKey();
		pedigree.add(key);
		
		while (key.getParent() != null) {
			key = key.getParent();
			pedigree.add(key);
		}
		
		Collections.reverse(pedigree);
		return pedigree;
	}
	
}
